from esxi_utils.vm.tools.guesttools import GuestTools, GuestToolsResponsePromise
from esxi_utils.util.response import Response
from esxi_utils.util import log, exceptions
import typing
import base64
import ntpath
import json
import os
import re

class WindowsGuestTools(GuestTools):
	"""
	Guest tools subclass specifically for interaction with a Windows operating system.

	Wrapper class for functionality related to VMware Tools on a Virtual Machine. The virtual machine must have VMware tools installed for
	any of the contained functions to work.

	:param vm: A `VirtualMachine` object to wrap. VMware tools must be installed on this virtual machine.
	"""
	def powershell_async(
		self, 
		username: str,
		password: str, 
		command: str, 
		cwd: typing.Optional[str] = None,
		powershell_path: str = "C:\\Windows\\System32\\WindowsPowerShell\\v1.0\\powershell.exe",
		out_stream_callback: typing.Optional[typing.Callable[[str], None]] = None
	):
		"""
		Runs a script in the guest operating system asynchronously using powershell.

		Unlike ``execute_program``, the command is executed as a standard powershell script and thus the full path to programs do not need to be provided. 
		Systems without powershell will fail. ``command`` may also be provided as a script, with newlines separating the commands to be run.

		:param username: The username of the virtual machine user.
		:param password: The password of the virtual machine user.
		:param command: The command or script to run.
		:param cwd: The working directory that the command should be run in.
		:param powershell_path: The path to the powershell executable.
		:param out_stream_callback: Optional callback to handle the stdout of commands. If not `None`, this will be called as a command's stdout becomes available. 
	
		:return: A `GuestToolsPowershellResponsePromise` object
		"""
		return GuestToolsPowershellResponsePromise(
			tools=self,
			username=username, 
			password=password, 
			command=command, 
			cwd=cwd,
			powershell_path=powershell_path, 
			out_stream_callback=out_stream_callback
		)

	def powershell(
		self, 
		username: str,
		password: str, 
		command: str, 
		cwd: typing.Optional[str] = None,
		powershell_path: str = "C:\\Windows\\System32\\WindowsPowerShell\\v1.0\\powershell.exe",
		timeout: typing.Optional[int] = 120,
		assert_status: typing.Optional[int] = None,
		out_stream_callback: typing.Optional[typing.Callable[[str], None]] = None
	) -> 'Response':
		"""
		Runs a script in the guest operating system using powershell. Unlike ``powershell_async``, this will block until the associated command has finished.

		Unlike ``execute_program``, the command is executed as a standard powershell script and thus the full path to programs do not need to be provided. 
		Systems without powershell will fail. ``command`` may also be provided as a script, with newlines separating the commands to be run.

		:param username: The username of the virtual machine user.
		:param password: The password of the virtual machine user.
		:param command: The command or script to run.
		:param cwd: The working directory that the command should be run in.
		:param powershell_path: The path to the powershell executable.
		:param timeout: The command timeout in seconds. Set to 0 or `None` to disable timeout.
		:param assert_status: Assert that the command exits with a certain status number. If `None`, no assertion is made and checking the status is left to the caller.
		:param out_stream_callback: Optional callback to handle the stdout of commands. If not `None`, this will be called as a command's stdout becomes available. 

		:return: A `Response` object
		"""
		return self.powershell_async(
			username=username, 
			password=password, 
			command=command, 
			cwd=cwd,
			powershell_path=powershell_path, 
			out_stream_callback=out_stream_callback
		).wait(assert_status=assert_status, timeout=timeout)

	def stat(
		self, 
		path: str, 
		username: str, 
		password: str, 
		powershell_path: str = "C:\\Windows\\System32\\WindowsPowerShell\\v1.0\\powershell.exe"
	) -> typing.Union[typing.Dict[str, typing.Any], None]:
		"""
		Get information about a file. PowerShell must be available on the system.

		Returned date formats in are seconds-since-Epoch.

		:param path: A path to a file or directory on the remote system.
		:param username: The username of the virtual machine user.
		:param password: The password of the virtual machine user.
		:param powershell_path: The path to the powershell executable.

		:return: A dict containing information about the file, or `None` if the file does not exist.
		"""
		path = ntpath.normpath(path)
		log.debug(f"{str(self)} Stat {path}")
		response = self.powershell(
			username=username,
			password=password,
			command=f"Get-Item {path} | ConvertTo-JSON",
			powershell_path=powershell_path
		)
		if response.status != 0 and (f"ObjectNotFound" in str(response.stderr) or f"ObjectNotFound" in str(response.stdout)):
			return None
		elif response.status != 0:
			raise exceptions.RemoteConnectionCommandError(self, response)
		
		results = json.loads(str(response.stdout).strip())

		def parse_date(datestr: str) -> float:
			date_match = re.search(r"Date\((\d+)\)", datestr)
			if not date_match:
				return -1
			date_value = date_match.group(1)
			if len(date_value) == 13:
				# In milliseconds
				return int(date_value) / 1000
			return int(date_value)

		return {
			"ProvidedPath": path,
			"Name": results["Name"],
			"FullName": results["FullName"],
			"Length": results.get("Length", 0),
			"IsFile": not str(results["Mode"]).startswith("d"),
			"IsDirectory": str(results["Mode"]).startswith("d"),
			"Mode": results["Mode"],
			"CreationTimeUtc": parse_date(results["CreationTimeUtc"]),
			"LastAccessTimeUtc": parse_date(results["LastAccessTimeUtc"]),
			"LastWriteTimeUtc": parse_date(results["LastWriteTimeUtc"]),
		}
	
	def bulkstat(self, paths: typing.List[str], username: str, password: str, powershell_path: str = "C:\\Windows\\System32\\WindowsPowerShell\\v1.0\\powershell.exe") -> typing.List[typing.Dict[str, typing.Any]]:
		"""
		Get information about one or more files.

		:param paths: The paths to the files or directories on the remote system.
		:param username: The username of the virtual machine user.
		:param password: The password of the virtual machine user.
		:param bash_path: The path to the bash executable.
		:param powershell_path: The path to the powershell executable.

		:return: A dict containing information about the file, or `None` if the file does not exist.
		"""
		paths = [ ntpath.normpath(p) for p in paths ]
		log.debug(f"{str(self)} Bulk Stat: {', '.join(paths)}")

		with self.use_temporary_file(username=username, password=password, extension=".ps1") as script_file:
			with self.use_temporary_file(username=username, password=password, extension=".stat") as output_file:
				script = f"""
$filenames = @({str(paths)[1:-1]})
foreach ($filename in $filenames) {{
    if (Test-Path $filename) {{
        Get-Item $filename | ConvertTo-Json -Compress | Out-File -Append -FilePath {output_file}
    }} else {{
        "Does not exist" | Out-File -Append -FilePath {output_file}
    }}
}}
""".strip()
				self.write_file(username=username, password=password, filepath=script_file, data=script)
				self.powershell(
					username=username,
					password=password,
					command=f"powershell -File {script_file}",
					powershell_path=powershell_path,
					assert_status=0
				)

				output: str = str(self.get_file(username=username, password=password, filepath=output_file, encoding="utf-16"))

		lines = output.strip().split("\n")
		if len(lines) != len(paths):
			raise RuntimeError(f"bulkstat: the number of output lines do not match the number of input files ({len(lines)} vs {len(paths)}). Raw Output:\n{output}")

		def parse_date(datestr: str) -> float:
			date_match = re.search(r"Date\((\d+)\)", datestr)
			if not date_match:
				return -1
			date_value = date_match.group(1)
			if len(date_value) == 13:
				# In milliseconds
				return int(date_value) / 1000
			return int(date_value)

		stats = list()
		for path, line in zip(paths, lines):
			if line.strip() == "Does not exist":
				stats.append({
					"ProvidedPath": path,
					"Exists": False,
					"Name": None,
					"FullName": None,
					"Length": None,
					"IsFile": None,
					"IsDirectory": None,
					"Mode": None,
					"CreationTimeUtc": None,
					"LastAccessTimeUtc": None,
					"LastWriteTimeUtc": None,
				})
				continue

			results = json.loads(line)
			stats.append({
				"ProvidedPath": path,
				"Exists": True,
				"Name": results["Name"],
				"FullName": results["FullName"],
				"Length": results.get("Length", 0),
				"IsFile": not str(results["Mode"]).startswith("d"),
				"IsDirectory": str(results["Mode"]).startswith("d"),
				"Mode": results["Mode"],
				"CreationTimeUtc": parse_date(results["CreationTimeUtc"]),
				"LastAccessTimeUtc": parse_date(results["LastAccessTimeUtc"]),
				"LastWriteTimeUtc": parse_date(results["LastWriteTimeUtc"]),
			})
		return stats

	def isfile(self, path: str, username: str, password: str, powershell_path: str = "C:\\Windows\\System32\\WindowsPowerShell\\v1.0\\powershell.exe") -> bool:
		"""
		Checks if the file exists and is a regular file.

		:param path: A path to a file on the remote system.
		:param username: The username of the virtual machine user.
		:param password: The password of the virtual machine user.
		:param powershell_path: The path to the powershell executable.
		
		:return: A boolean whether or not the file exists.
		"""
		path = ntpath.normpath(path)
		stat = self.stat(
			path, 
			username=username, 
			password=password,
			powershell_path=powershell_path
		)
		if stat is None:
			return False
		return stat["IsFile"]

	def isdir(self, path: str, username: str, password: str, powershell_path: str = "C:\\Windows\\System32\\WindowsPowerShell\\v1.0\\powershell.exe") -> bool:
		"""
		Checks if the directory exists.

		:param path: A path to a directory on the remote system.
		:param username: The username of the virtual machine user.
		:param password: The password of the virtual machine user.
		:param powershell_path: The path to the powershell executable.
		
		:return: A boolean whether or not the directory exists.
		"""
		path = ntpath.normpath(path)
		stat = self.stat(
			path, 
			username=username, 
			password=password,
			powershell_path=powershell_path
		)
		if stat is None:
			return False
		return stat["IsDirectory"]

	def download(
		self, 
		path: str, 
		dst: str, 
		username: str,
		password: str,
		directory_contents_only: bool = False, 
		overwrite: bool = False,
		powershell_path: str = "C:\\Windows\\System32\\WindowsPowerShell\\v1.0\\powershell.exe"
	) -> typing.List[str]:
		"""
		Download a file or directory from the remote system.

		:param path: Path to the file or directory on the remote system.
		:param dst: The local destination file or directory to download to.
		:param username: The username of the virtual machine user.
		:param password: The password of the virtual machine user.
		:param directory_contents_only: If `True` and `path` points to a directory, then just the contents of the directory will be downloaded rather than the directory itself.
		:param overwrite: Whether to overwrite existing files.
		:param powershell_path: The path to the powershell executable.

		:return: A list of local paths to all files downloaded.
		"""
		path = ntpath.normpath(path)
		log.info(f"{str(self)} Downloading \"{path}\" to \"{dst}\"...")
		dst = os.path.abspath(dst)
		basename = ntpath.basename(path)
		stat = self.stat(path, username=username, password=password, powershell_path=powershell_path)
		if stat is None:
			raise exceptions.RemoteFileNotFoundError(self, path)

		download_map: typing.Dict[str, str] = {} # Map download source -> destination
		if stat["IsFile"]:
			# Single file download
			target = dst
			if not os.path.exists(target):
				# Target doesn't exist, we will download under the name specified
				# Just ensure the parent directory does exist
				parent = os.path.dirname(target)
				if not os.path.isdir(parent):
					raise NotADirectoryError(f"{parent} is not a valid local directory")
			else:
				# Target exists, we'll download into the target with the same name
				if not os.path.isdir(target):
					raise NotADirectoryError(f"{target} is not a valid local directory")
				target = os.path.join(target, basename)
			download_map[path] = target
		else:
			# Directory download
			target = dst
			if os.path.isdir(target) and directory_contents_only:
				# Downloading directory contents into a directory
				pass
			elif os.path.isdir(target) and not directory_contents_only:
				# Downloading directory into a directory, download with the same name into the target
				target = os.path.join(target, stat["Name"])
			elif not os.path.exists(target) and directory_contents_only:
				# We cannot download the contents of a directory if the dst does not exist
				raise NotADirectoryError(f"{dst} is not a valid local directory")
			elif not os.path.exists(target) and not directory_contents_only:
				# Downloading directory under a new name, we just need to ensure that the parent directory exists
				parent = os.path.dirname(target)
				if not os.path.isdir(parent):
					raise NotADirectoryError(f"{parent} is not a valid local directory")
			
			download_map = {}
			files_to_download = [ ntpath.join(path, filename) for filename in self.list_files(username=username, password=password, filepath=path) ]
			while len(files_to_download):
				stats = self.bulkstat(files_to_download, username=username, password=password, powershell_path=powershell_path)
				files_to_download = []
				for stat in stats:
					if stat["IsDirectory"]:
						files_to_download.extend([ ntpath.join(stat["FullName"], filename) for filename in self.list_files(username=username, password=password, filepath=stat["FullName"]) ])
						continue
					download_map[stat["FullName"]] = os.path.join(target, stat["FullName"][len(path):].replace("\\", "/").strip("/"))

		if len(download_map.keys()) == 0:
			return []

		for target in download_map.values():
			if os.path.exists(target) and not overwrite:
				raise FileExistsError(f"{target} already exists")
			if os.path.isdir(target):
				raise IsADirectoryError(f"{target} cannot be overwritten as it exists as a directory")

		# Begin downloading files
		for remotepath, localpath in download_map.items():
			log.debug(f"{self} Downloading {remotepath} to {localpath}")
			os.makedirs(os.path.dirname(localpath), exist_ok=True)
			file_bytes: bytes = self.get_file(username=username, password=password, filepath=remotepath, encoding=None) # type: ignore
			with open(localpath, mode="wb") as f:
				f.write(file_bytes)

		return list(download_map.values())

	def upload(
		self, 
		src: str,
		dst: str,
		username: str,
		password: str,
		directory_contents_only: bool = False, 
		overwrite: bool = False,
		powershell_path: str = "C:\\Windows\\System32\\WindowsPowerShell\\v1.0\\powershell.exe"
	) -> typing.List[str]:
		"""
		Upload a local file or directory to this path on the remote system.
		
		Warning: This function will load entire files into memory. Ensure that extremely large files are not uploaded using this method.

		:param src: The local path to a file or directory to upload.
		:param dst: The remote path to upload to.
		:param username: The username of the virtual machine user.
		:param password: The password of the virtual machine user.
		:param directory_contents_only: If `True` and `src` points to a directory, then only the contents of the directory will be uploaded rather than the directory itself.
		:param overwrite: Whether to overwrite existing files on the remote system.
		:param powershell_path: The path to the powershell executable.

		:return: A list of remote paths to all files uploaded.
		"""
		dst = ntpath.normpath(dst)
		log.info(f"<{self}> Uploading \"{src}\" to \"{dst}\"...")
		src = os.path.abspath(src)
		if not os.path.exists(src):
			raise FileNotFoundError(f"File or directory {src} not found")

		dst_stat = self.stat(dst, username=username, password=password, powershell_path=powershell_path)

		upload_map: typing.Dict[str, str] = {} # Map upload source -> destination
		if os.path.isfile(src):
			# Single file upload
			target = dst
			if dst_stat is None:
				# Target doesn't exist, we will upload under the name specified
				# Just ensure the parent directory does exist
				parent = ntpath.dirname(target)
				if not self.isdir(parent, username=username, password=password, powershell_path=powershell_path):
					raise exceptions.RemoteNotADirectoryError(self, parent)
			else:
				# Target exists, we'll upload into the target with the same name
				if not dst_stat["IsDirectory"]:
					raise exceptions.RemoteNotADirectoryError(self, dst)
				target = ntpath.join(target, os.path.basename(src))
			upload_map[src] = target
		elif os.path.isdir(src):
			# Directory upload
			target = dst
			target_exists = dst_stat is not None
			target_is_dir = dst_stat and dst_stat["IsDirectory"]
			if target_is_dir and directory_contents_only:
				# Uploading directory contents into a directory
				pass
			elif target_is_dir and not directory_contents_only:
				# Uploading directory into a directory, upload with the same name into the target
				target = ntpath.join(target, os.path.basename(src))
			elif not target_exists and directory_contents_only:
				# We cannot upload the contents of a directory if the dst does not exist
				raise exceptions.RemoteNotADirectoryError(self, target)
			elif not target_exists and not directory_contents_only:
				# Uploading directory under a new name, we just need to ensure that the parent directory exists
				parent = ntpath.dirname(target)
				if not self.isdir(parent, username=username, password=password, powershell_path=powershell_path):
					raise exceptions.RemoteNotADirectoryError(self, parent)
			
			for root, _, files in os.walk(src):
				for file in files:
					path = os.path.join(root, file)
					upload_map[path] = ntpath.join(target, ntpath.normpath(path[len(src):]).strip("\\"))

		if len(upload_map.keys()) == 0:
			return []

		for stat in self.bulkstat(list(upload_map.values()), username=username, password=password, powershell_path=powershell_path):
			if stat["Exists"] and stat["IsFile"] and not overwrite:
				raise exceptions.RemoteFileExistsError(self, stat["ProvidedPath"])
			if stat["Exists"] and stat["IsDirectory"]:
				raise exceptions.RemoteIsADirectoryError(self, stat["ProvidedPath"])

		# Determine the directories to create
		# We will be running mkdir with parents, so we only want the longest paths
		dirs_to_make = []
		for remotepath in upload_map.values():
			dirname = ntpath.dirname(remotepath).rstrip("\\") + "\\"
			dirs_to_make = [ path for path in dirs_to_make if not dirname.startswith(path) ] # Remove any paths that are subsets of this path 
			is_subpath = any([ path.startswith(dirname) for path in dirs_to_make ]) # Check if this is a subpath of any existing path already
			if not is_subpath:
				dirs_to_make.append(dirname)

		# Make directories
		self.powershell(
			username=username, 
			password=password,
			command=f"mkdir {str(dirs_to_make)[1:-1]} -Force -ea 0",
			powershell_path=powershell_path,
			assert_status=0
		)

		# Begin uploading files
		for localpath, remotepath in upload_map.items():
			log.debug(f"<{self}> Uploading \"{localpath}\" to \"{remotepath}\"")
			with open(localpath, "rb") as f:
				self.write_file(username=username, password=password, filepath=remotepath, data=f.read())

		return list(upload_map.values())



class GuestToolsPowershellResponsePromise(GuestToolsResponsePromise):
	"""
	Encapsulates a Promise for a result of an GuestTools Powershell command.

	:param tools: The WindowsGuestTools instance from which this promise originates.
	:param username: The username of the virtual machine user.
	:param password: The password of the virtual machine user.
	:param command: The command or script to run.
	:param cwd: The working directory that the command should be run in.
	:param powershell_path: The path to the powershell executable.
	:param out_stream_callback: Optional callback to handle the stdout of commands. If not `None`, this will be called as a command's stdout becomes available. 
	"""
	def __init__(
		self, 
		tools: WindowsGuestTools, 
		username: str,
		password: str, 
		command: str, 
		cwd: typing.Optional[str] = None,
		powershell_path: str = "C:\\Windows\\System32\\WindowsPowerShell\\v1.0\\powershell.exe",
		out_stream_callback: typing.Optional[typing.Callable[[str], None]] = None
	):
		assert isinstance(powershell_path, str), "powershell_path must be a string"

		self._powershell_path = powershell_path
		self._exitcodefile = None

		super().__init__(
			tools=tools, 
			username=username, 
			password=password, 
			command=command,
			cwd=cwd,
			make_output_available=True,
			out_stream_callback=out_stream_callback
		)

	def _encode_command(self, command_string: str) -> str:
		"""
		Base64 encode the given command so that it can be passed to powershell.

		:param command_string: The command to encode.
		
		:return: The base64 encoded command.
		"""
		return base64.b64encode(str.encode(command_string, encoding="utf_16_le")).decode()

	def _setup(self, make_output_available: bool = False) -> str:
		"""
		Perform any setup necessary prior to starting the command.

		:param make_output_available: Whether or not it is requested to make the output of the command available.
		
		:return: The command to run.
		"""
		self._exitcodefile = self._tools.create_temporary_file(self._username, self._password, ".log")
		encoded_command = self._encode_command(self._command)

		run_script = ""
		if make_output_available:
			self._stdoutfile = self._tools.create_temporary_file(self._username, self._password, ".out")
			self._stderrfile = self._tools.create_temporary_file(self._username, self._password, ".err")

			run_script = f"""
			$process = Start-Process "powershell.exe" -ArgumentList "-NonInteractive -EncodedCommand {encoded_command}" -RedirectStandardError "{self._stderrfile}" -RedirectStandardOutput "{self._stdoutfile}" -PassThru -NoNewWindow -Wait
			$process.ExitCode | Out-File -FilePath "{self._exitcodefile}"
			"""
		else:
			run_script = f"""
			$process = Start-Process "powershell.exe" -ArgumentList "-NonInteractive -EncodedCommand {encoded_command}" -PassThru -NoNewWindow -Wait
			$process.ExitCode | Out-File -FilePath "{self._exitcodefile}"
			"""
		run_script = "\r\n".join([ line.strip() for line in run_script.replace("\r", "").split("\n") ])

		exec_command = f"""Start-Process "{self._powershell_path}" -ArgumentList "-NonInteractive -EncodedCommand {self._encode_command(run_script)}" -NoNewWindow -Wait"""
		return f"{self._powershell_path} -EncodedCommand {self._encode_command(exec_command)}" 

	def _format_response(self, response: Response) -> Response:
		"""
		Perform post-processing on a result. This will occur when a command finishes executing but before any files are cleaned up.

		:param response: The result of the command.

		:return: The new result.
		"""
		if response.status == 0:
			try:
				assert self._exitcodefile, "No exitcode file"
				status_code = self._tools.get_file(self._username, self._password, self._exitcodefile, encoding="utf-16").strip()
				response.status = int(status_code)
			except ValueError as e:
				print(e)
				response.status = 1
		return response

	def _cleanup(self):
		"""
		Clean up any files or other resources after the process has finished.
		"""
		super()._cleanup()
		if self._exitcodefile:
			self._tools.delete_file(self._username, self._password, self._exitcodefile)
			self._exitcodefile = None



from esxi_utils.vm.hardware.device import VirtualDevice, VirtualDeviceList
from esxi_utils.util import log, exceptions, parse
from esxi_utils.datastore import DatastoreFile
import pyVmomi
import typing
import re

class VirtualDiskList(VirtualDeviceList):
	"""
	The list of all virtual disks on a Virtual Machine.
	"""
	def __iter__(self) -> typing.Iterator['VirtualDisk']:
		disks = [ dev for dev in super().__iter__() if isinstance(dev, VirtualDisk) ]
		disks.sort(key=lambda x: int(re.search(r"\d+", x.label).group(0)))
		return iter(disks)

	def __getitem__(self, index) -> 'VirtualDisk':
		return self.items[index]

	@property
	def items(self) -> 'VirtualDisk':
		"""
		A list of all items
		"""
		return list(self)

	def add(self, size: typing.Union[str, int], scsi: int = 0, unit_number: typing.Optional[int] = None, thin: bool = True) -> 'VirtualDisk':
		"""
		Adds a disk to this VM.

		:param size:
			A string or integer representing the size of the new disk.
			If this is an integer, this represents a size in KB.
			If this is a string, the string is expected to take the form <number><unit>, where <unit> is one of KB, MB, or GB (e.g. 32GB).
		:param scsi:
			The number for the SCSI controller to attach the new disk to.
		:param unit_number:
			The unit number to assign the disk to on the provided controller. If `None`, the next available unit number will be used.
		:param thin:
			Whether or not this new disk should be thin provisioned.

		:return: The `VirtualDisk` object for the newly added disk.
		"""
		log.info(f"{str(self)} Adding new disk with size = {size}, scsi = {scsi}, unit number = {unit_number}, thin provisioning = {thin}")
		if isinstance(size, str):
			size = parse.size_string(size, unit="KB")
		if self._vm.snapshots.exists:
			raise exceptions.SnapshotsExistError(self._vm)
		
		# Find the requested SCSI controller
		scsi_controller = None
		for device in super().__iter__():
			if device.label == f"SCSI controller {scsi}":
				scsi_controller = device
				break
		else:
			raise exceptions.VirtualMachineHardwareNotFoundError(self._vm, "SCSI Controller", f"No SCSI {scsi} found")

		# Get the unit number if needed
		if unit_number is None:
			scsi_devices = scsi_controller._obj.device if scsi_controller._obj.device else []
			scsi_disks = [ disk for disk in self if disk.key in scsi_devices ]
			taken_unit_numbers = [ disk._obj.unitNumber for disk in scsi_disks ]
			available_unit_numbers = [ i for i in range(16) if i not in taken_unit_numbers ]
			if len(available_unit_numbers) == 0:
				raise exceptions.VirtualMachineHardwareError(self._vm, f"Failed to determine unit number for SCSI {scsi}")
			log.debug(f"{str(self)} Found next unit number for SCSI={scsi}: {available_unit_numbers[0]}")
			unit_number = available_unit_numbers[0]

		if not isinstance(unit_number, int) or unit_number < 0:
			raise ValueError(f"unit_number must be an integer greater or equal to 0")

		disk_spec = pyVmomi.vim.vm.device.VirtualDeviceSpec()
		disk_spec.fileOperation = "create"
		disk_spec.operation = pyVmomi.vim.vm.device.VirtualDeviceSpec.Operation.add
		disk_spec.device = pyVmomi.vim.vm.device.VirtualDisk()
		disk_spec.device.backing = pyVmomi.vim.vm.device.VirtualDisk.FlatVer2BackingInfo()
		if thin:
			disk_spec.device.backing.thinProvisioned = True
		disk_spec.device.backing.diskMode = 'persistent'
		disk_spec.device.unitNumber = unit_number
		disk_spec.device.capacityInKB = size
		disk_spec.device.controllerKey = scsi_controller.key
		return self._add_device(disk_spec)

	def add_existing_disk(self, filepath: typing.Optional['DatastoreFile'] = None, scsi: int = 0, unit_number: typing.Optional[int] = None) -> 'VirtualDisk':
		"""
		Adds an existing disk to this VM using a datastore file path.

		:param datastore_file_path:
			The full datastore file path to the existing virtual disk (e.g., '[datastore1] folder/disk.vmdk').
		:param scsi:
			The number for the SCSI controller to attach the disk to.
		:param unit_number:
			The unit number to assign the disk to on the provided controller. If `None`, the next available unit number will be used.

		:return: The `VirtualDisk` object for the newly added disk.
		"""
		assert isinstance(filepath, DatastoreFile) or filepath is None, "filepath must be a DatastoreFile object or None"
		log.info(f"{str(self)} Adding existing disk with datastore file path = {filepath}, scsi = {scsi}, unit number = {unit_number}")
		if self._vm.snapshots.exists:
			raise exceptions.SnapshotsExistError(self._vm)

		# Find the requested SCSI controller
		scsi_controller = None
		for device in super().__iter__():
			if device.label == f"SCSI controller {scsi}":
				scsi_controller = device
				break
		else:
			raise exceptions.VirtualMachineHardwareNotFoundError(self._vm, "SCSI Controller", f"No SCSI {scsi} found")

		# Get the unit number if needed
		if unit_number is None:
			scsi_devices = scsi_controller._obj.device if scsi_controller._obj.device else []
			scsi_disks = [disk for disk in self if disk.key in scsi_devices]
			taken_unit_numbers = [disk._obj.unitNumber for disk in scsi_disks]
			available_unit_numbers = [i for i in range(16) if i not in taken_unit_numbers]
			if len(available_unit_numbers) == 0:
				raise exceptions.VirtualMachineHardwareError(self._vm, f"Failed to determine unit number for SCSI {scsi}")
			log.debug(f"{str(self)} Found next unit number for SCSI={scsi}: {available_unit_numbers[0]}")
			unit_number = available_unit_numbers[0]

		if not isinstance(unit_number, int) or unit_number < 0:
			raise ValueError(f"unit_number must be an integer greater or equal to 0")

		# Create the disk specification for the existing disk
		disk_spec = pyVmomi.vim.vm.device.VirtualDeviceSpec()
		disk_spec.operation = pyVmomi.vim.vm.device.VirtualDeviceSpec.Operation.add
		disk_spec.device = pyVmomi.vim.vm.device.VirtualDisk()
		disk_spec.device.backing = pyVmomi.vim.vm.device.VirtualDisk.FlatVer2BackingInfo()
		disk_spec.device.backing.fileName = filepath.path
		disk_spec.device.backing.diskMode = 'persistent'
		disk_spec.device.unitNumber = unit_number
		disk_spec.device.controllerKey = scsi_controller.key

		return self._add_device(disk_spec)


class VirtualDisk(VirtualDevice):
	@property
	def size(self) -> int:
		"""
		The size of this disk in KB.
		"""
		return self._obj.capacityInKB

	@size.setter
	def size(self, size : typing.Union[str, int]):
		"""
		Modify the size of this disk.

		:param size:
			A string or integer representing the new size of this disk. The size must be greater or equal to the current disk size.
			If this is an integer, this represents a size in KB.
			If this is a string, the string is expected to take the form <number><unit>, where <unit> is one of KB, MB, or GB (e.g. 32GB).
		"""
		if isinstance(size, str):
			size = parse.size_string(size, unit="KB")
		log.info(f"{str(self)} Updating size to: {self.size}KB")

		if self._vm.snapshots.exists:
			raise exceptions.SnapshotsExistError(self._vm)
		if size <= self.size:
			raise exceptions.VirtualMachineInvalidHardwareConfigurationError(self._vm, f"specified disk size ({size}KB) must be greater than the existing disk size ({self.size}KB)")

		device_spec = pyVmomi.vim.vm.device.VirtualDeviceSpec()
		device_spec.operation = pyVmomi.vim.vm.device.VirtualDeviceSpec.Operation.edit
		device_spec.device = self._obj
		device_spec.device.capacityInKB = size
		spec = pyVmomi.vim.vm.ConfigSpec()
		spec.deviceChange = [device_spec]
		self._vm._client._wait_for_task(
			self._vm._vim_vm.ReconfigVM_Task(spec=spec)
		)

	@property
	def filepath(self) -> 'DatastoreFile':
		"""
		The path to this disk's file as a `DatastoreFile` object.
		"""
		return self._get_backing_file()

	def remove(self, delete_file: bool = True):
		"""
		Remove this disk from this virtual machine.

		:param delete_file: Delete the file from the datastore.
		"""
		log.info(f"{str(self)} Removing device")
		if self._vm.snapshots.exists:
			raise exceptions.SnapshotsExistError(self._vm)
		device_spec = pyVmomi.vim.vm.device.VirtualDeviceSpec()
		device_spec.operation = pyVmomi.vim.vm.device.VirtualDeviceSpec.Operation.remove
		if delete_file:
			device_spec.fileOperation = pyVmomi.vim.vm.device.VirtualDeviceSpec.FileOperation.destroy
		device_spec.device = self._obj
		spec = pyVmomi.vim.vm.ConfigSpec()
		spec.deviceChange = [device_spec]
		self._vm._client._wait_for_task(
			self._vm._vim_vm.ReconfigVM_Task(spec=spec)
		)

	def __str__(self):
		return f"<{type(self).__name__}(size={self.size}KB) for VM='{self._vm.name}'>"


from esxi_utils.util import log, exceptions
import pyVmomi
import typing

if typing.TYPE_CHECKING:
	from esxi_utils.client import ESXiClient
	from esxi_utils.vm.virtualmachine import VirtualMachine
	from esxi_utils.networking.distributedvswitch import DistributedVSwitch
	# from esxi_utils.networking.vmkernelnic import VMKernelNIC

class DistributedPortGroupList:
	"""
	The list of port groups on the ESXi host.
	"""
	def __init__(self, client: 'ESXiClient'):
		self._client = client

	def __iter__(self) -> typing.Iterator['DistributedPortGroup']:
		return iter([ DistributedPortGroup(self._client, pg.portgroupName) for pg in self._distributed_system ])

	def __getitem__(self, name) -> 'DistributedPortGroup':
		return self.get(name)

	def __contains__(self, name: str) -> bool:
		return self.exists(name)

	@property
	def items(self) -> typing.List['DistributedPortGroup']:
		"""
		A list of all items
		"""
		return list(self)

	def find(self, name: str) -> typing.Union['DistributedPortGroup', None]:
		"""
		Get a port group by name.

		:param name: The name of the port group.

		:return: A `PortGroup` object, or `None` if not found.
		"""
		found = [ pg for pg in self if pg.name == name ]
		if len(found) != 1:
			return None
		return found[0]

	def get(self, name: str) -> 'DistributedPortGroup':
		"""
		Get a port group by name and raise an exception if not found.

		:param name: The name of the port group.

		:return: A `PortGroup` object
		"""
		pg = self.find(name)
		if pg is None:
			raise exceptions.NetworkingObjectNotFoundError('PortGroup', name)
		return pg

	def exists(self, name: str) -> bool:
		"""
		Check whether a port group exists 

		:param name: The name of the port group

		:return: Whether or not the port group exists
		"""
		return self.find(name) is not None

	@property
	def _distributed_system(self):
		return self._client._host_system.parent.environmentBrowser.QueryConfigTarget().distributedVirtualPortgroup

	def __str__(self):
		return f"<{type(self).__name__} ({len(list(self))} Distributed Port Groups)>"
	
	def __repr__(self):
		return str(self)


class DistributedPortGroup:
	"""
	A Portgroup on the ESXi host.
	"""
	def __init__(self, client: 'ESXiClient', name: str):
		self._client = client
		self._name = name

	@property
	def name(self) -> str:
		"""
		The name of the port group.
		"""
		return self._name

	@property
	def vlan(self) -> int:
		"""
		The VLAN ID for ports using this port group. Possible values:
		- A value of 0 specifies that you do not want the port group associated with a VLAN.
		- A value from 1 to 4094 specifies a VLAN ID for the port group.
		- A value of 4095 specifies that the port group should use trunk mode, which allows the guest operating system to manage its own VLAN tags. 
		- A 'vim.NumericRange' object representing the range of available ports (contains a 'start' and an 'end' value to mark the range)
		"""
		return self._obj.portgroup.config.defaultPortConfig.vlan.vlanId

	# @property
	# def ports(self) -> typing.List[typing.Dict[str, str]]:
	# 	"""
	# 	The ports that currently exist and are used on this port group.
	# 	Returns a list of objects with the following keys:
	# 	- mac: The Media Access Control (MAC) address of network service of the virtual machine connected on this port. 
	# 	- type: The type of component connected on this port. One of:
	# 		- host: The VMkernel is connected to this port group. 
	# 		- systemManagement: A system management entity (service console) is connected to this port group. 
	# 		- unknown: This port group serves an entity of unspecified kind. 
	# 		- virtualMachine: A virtual machine is connected to this port group. 
	# 	"""
	# 	return [ { "mac": port.mac[0] if len(port.mac) else None, "type": port.type } for port in self._obj.port ]

	# @property
	# def active_clients(self) -> int:
	# 	"""
	# 	The number of active clients of this port group (the number of connections to powered-on virtual machines).
	# 	"""
	# 	return len([ port for port in self.ports if port["type"] == "virtualMachine" ])

	@property
	def vswitch(self) -> 'DistributedVSwitch':
		"""
		The Distributed VSwitch that this distributed port group belongs to.
		"""
		return self.distributed_vswitch

	@property
	def distributed_vswitch(self) -> 'DistributedVSwitch':
		"""
		The Distributed VSwitch that this distributed port group belongs to.
		"""
		from esxi_utils.networking.distributedvswitch import DistributedVSwitch
		return DistributedVSwitch(self._client, self._obj.switchName)

	# @property
	# def vmkernelnic(self) -> 'VMKernelNIC':
	# 	"""
	# 	The VMKernel NICs assigned to this port group, if any.
	# 	"""
	# 	host_ports = [ port for port in self.ports if port["type"] == "host" ]
	# 	if len(host_ports) == 0:
	# 		return None
	# 	if len(host_ports) > 1:
	# 		# Can there be multiple VMKernel NICs assigned to a port group?
	# 		# ESXi appears to only allow one, but this is uncertain so we log a warning here and future support
	# 		# for multiple may be added if a situation arises where this is not the case
	# 		log.warn(f"Mutiple host ports are assigned to {str(self)} but only one VMKernel NICs is supported.")
	# 	found = [ vnic for vnic in self._client.vmkernelnics if vnic.mac == host_ports[0]["mac"] ]
	# 	if len(found) == 0:
	# 		raise exceptions.NetworkingError(self, "No VMKernel NICs found")
	# 	return found[0]

	@property
	def vms(self) -> typing.List['VirtualMachine']:
		"""
		A list of the virtual machines attached to this distributed port group.
		"""
		vms = []
		for vm in self._client.vms:
			for nic in vm.nics:
				if nic.network == self.name:
					vms.append(vm)
					break
		return vms

	# def remove(self):
	# 	"""
	# 	Remove this port group from the system.
	# 	"""
	# 	self._network_system.RemovePortGroup(self.name)

	@property
	def _obj(self):
		for pg in self._distributed_system:
			if pg.portgroupName == self.name:
				return pg
		raise exceptions.NetworkingObjectNotFoundError('DistributedPortGroup', self.name)

	@property
	def _distributed_system(self):
		return self._client._host_system.parent.environmentBrowser.QueryConfigTarget().distributedVirtualPortgroup

	def __str__(self):
		return f"<{type(self).__name__} \"{self.name}\" vlan={self.vlan} on DistributedVSwitch \"{self.distributed_vswitch.name}\">"
	
	def __repr__(self):
		return str(self)


from datetime import datetime
import typing
import re

def size_string(size: str, unit: str = "KB") -> int:
	"""
	Convert a size string (e.g. 10MB, 10GB, etc.) into a integer equivalent in `unit`.

	:param size: The size string.
	:param unit: The unit (KB, MB, or GB) that the size string should be converted to.

	:return: The size converted to an int in `unit` units.
	"""
	CONVERSION_FACTORS = {
		"KB": 1,
		"MB": 1024,
		"GB": 1024 * 1024
	}
	assert unit in CONVERSION_FACTORS, f"Not a valid unit, use one of: {', '.join(list(CONVERSION_FACTORS.keys()))}"
	size = size.upper()
	from_unit = None
	for u in list(CONVERSION_FACTORS.keys()):
		if size.endswith(u):
			from_unit = u
			break
	if from_unit is None:
		raise ValueError(f"Not a valid unit for 'size'. Use one of: {', '.join(list(CONVERSION_FACTORS.keys()))}")

	size = size[:-len(from_unit)]
	try:
		size = int(size)
		if size < 1:
			raise ValueError("size cannot be less than 1")
	except ValueError:
		raise ValueError("size could not be read as a positive integer")
		
	size = size * CONVERSION_FACTORS[from_unit] # Convert to KB first
	size = size / CONVERSION_FACTORS[unit] # Convert to target unit
	return int(size)


def vimobj(vimobj_string: str, include_dtype: bool = True) -> typing.Dict[str, typing.Any]:
	"""
	Parse a vim object string into a dictionary.

	:param vimobj_string: The string.
	:param include_dtype: Include object types alongside their data.

	:return: A dictionary representation of the vim object.
	"""
	def vimcmd_lexer(src):
		tokens = []
		chars = ""
		for i, char in enumerate(src): 
			chars += char
			if chars.endswith("\n") or chars.endswith(","):
				value = chars[:-1].strip()
				if value == "true":
					tokens.append({ "type": "literal", "contents": True })
					chars = ""
				if value == "false":
					tokens.append({ "type": "literal", "contents": False })
					chars = ""
				if value == "null" or value == "<unset>":
					tokens.append({ "type": "literal", "contents": None })
					chars = ""
				try:
					x = int(value)
					tokens.append({ "type": "literal", "contents": x })
					chars = ""
				except ValueError:
					pass

				if (value.startswith("\"") or value.startswith("\'")) and len(value) > 1 and value.endswith(value[0]):
					# Try to parse the string ending here
					# Use a backtracking algorithm to continue the string if it cannot parse the remaining after this point
					try:
						x = vimcmd_lexer(src[i+1:])
						value = value[1:-1]
						try:
							# Try to parse as date
							value = datetime.fromisoformat(value.rstrip("Z"))
						except ValueError:
							pass
						tokens.append({ "type": "literal", "contents": value })
						tokens.extend(x)
						return tokens
					except Exception as e:
						pass

				elif len(tokens) != 0 and tokens[-1]["type"] == "field":
					# Catch-all
					# Also use a backtracking algorithm to continue if it cannot parse the remaining after this point
					try:
						x = vimcmd_lexer(src[i+1:])
						tokens.append({ "type": "literal", "contents": str(value) })
						tokens.extend(x)
						return tokens
					except Exception as e:
						pass

			stripped = chars.strip()
			if stripped == ",":
				chars = ""
				continue
			if re.match(r"\([^\)]+\)", stripped):
				tokens.append({ "type": "obj_name", "contents": stripped[1:-1] })
				chars = ""
			if stripped == "{":
				tokens.append({ "type": "open_bracket", "contents": dict() })
				chars = ""
			if stripped == "}":
				tokens.append({ "type": "close_bracket", "contents": None })
				chars = ""
			if stripped == "[":
				tokens.append({ "type": "open_list", "contents": list() })
				chars = ""
			if stripped == "]":
				tokens.append({ "type": "close_list", "contents": None })
				chars = ""
			if re.match(r"\S+ =", stripped):
				tokens.append({ "type": "field", "contents": stripped.split("=")[0].strip() })
				chars = ""
		
		if len(chars.strip()) != 0:
			raise ValueError(f"Failed to lex vim-cmd string: {src}\nRemaining chars: {chars}\nTokens: {tokens}")
		return tokens

	def vimcmd_parse_next(tokens):
		if tokens[0]["type"] == "literal":
			return (tokens[0]["contents"], tokens[1:])

		if tokens[0]["type"] == "obj_name":
			if tokens[1]["type"] == "literal":
				if include_dtype:
					return ({ "dtype": tokens[0]["contents"], "value": tokens[1]["contents"] }, tokens[2:])
				else:
					return (tokens[1]["contents"], tokens[2:])
			elif tokens[1]["type"] in ["open_bracket", "open_list"]:
				start_token = tokens[1]["type"]
				obj = tokens[1]["contents"]
				end_token = "close_bracket" if start_token == "open_bracket" else "close_list"
				num_open = 1
				inner_tokens = [ ]
				i = 2
				while num_open != 0:
					if tokens[i]["type"] == start_token:
						num_open += 1
					elif tokens[i]["type"] == end_token:
						num_open -= 1
					if num_open != 0:
						inner_tokens.append(tokens[i])
					i += 1

				while len(inner_tokens) != 0:
					if inner_tokens[0]["type"] == "field":
						field_name = inner_tokens[0]["contents"]
						obj[field_name], inner_tokens = vimcmd_parse_next(inner_tokens[1:])
					else:
						val, inner_tokens = vimcmd_parse_next(inner_tokens)
						obj.append(val)

				if include_dtype:
					return ({ "dtype": tokens[0]["contents"], "value": obj }, tokens[i:])
				else:
					return (obj, tokens[i:])
			else:
				raise ValueError(f"Unknown token for object: {str(tokens[1])}")
		raise ValueError(f"Unknown token: {str(tokens[1])}")

	def vimcmd_parse(src):
		src = src.replace("\r", "")
		match = re.search(r"^\([^\)]+\)\s+[\[|\{]", src, flags=re.MULTILINE)
		if match is None:
			raise ValueError("String is unparsable.")
		src = src[match.start():]
		tokens = vimcmd_lexer(src)
		return vimcmd_parse_next(tokens)[0]

	try:
		return vimcmd_parse(vimobj_string)
	except Exception as e:
		raise ValueError(f"Error while parsing vim object: {str(e)}")


import typing

class NoOpLogger:
	def debug(self, msg): pass
	def info(self, msg): pass
	def warning(self, msg): pass
	def error(self, msg): pass
	def critical(self, msg): pass


Logger = NoOpLogger()
"""The logger instance to use for logging."""


def enable(handler: typing.Any):
	"""
	Enable logging using the given handler. This is duck-typed: the handler may be any object, as long as it implements the following function definitions::

		debug(str)
		info(str)
		warning(str)
		error(str)
		critical(str)

	:param handler: The handler to use.
	"""
	global Logger
	for method in [ "debug", "info", "warning", "error", "critical" ]:
		assert hasattr(handler, method) and callable(getattr(handler, method)), f"Handler does not implement method: {method}(str)"
	Logger = handler
	

def debug(msg: str):
	"""
	Write a debug message.

	:param msg: The message to write.
	"""
	global Logger
	Logger.debug(msg)


def info(msg: str):
	"""
	Write an info message.

	:param msg: The message to write.
	"""
	global Logger
	Logger.info(msg)


def warning(msg: str):
	"""
	Write a warning message.

	:param msg: The message to write.
	"""
	global Logger
	Logger.warning(msg)


def error(msg: str):
	"""
	Write an error message.

	:param msg: The message to write.
	"""
	global Logger
	Logger.error(msg)


def critical(msg: str):
	"""
	Write a critical message.

	:param msg: The message to write.
	"""
	global Logger
	Logger.critical(msg)
	


from esxi_utils.util import log, exceptions
from esxi_utils.datastore import DatastoreFile
import pyVmomi
import typing

if typing.TYPE_CHECKING:
	from esxi_utils.vm.virtualmachine import VirtualMachine

class VirtualDeviceList:
	"""
	The list of all virtual devices on a Virtual Machine.
	"""
	def __init__(self, vm: 'VirtualMachine'):
		self._vm = vm
		self._vim_vm = self._vm._vim_vm

	def __iter__(self) -> typing.Iterator['VirtualDevice']:
		from esxi_utils.vm.hardware.nic import VirtualNIC
		from esxi_utils.vm.hardware.disk import VirtualDisk
		from esxi_utils.vm.hardware.cdrom import VirtualCdrom
		from esxi_utils.vm.hardware.floppy import VirtualFloppy
		from esxi_utils.vm.hardware.video_card import VirtualVideoCard

		devices = []
		for obj in self._vim_vm.config.hardware.device:
			device = None
			key = obj.key
			if isinstance(obj, pyVmomi.vim.vm.device.VirtualEthernetCard):
				device = VirtualNIC(self._vm, key)
			elif isinstance(obj, pyVmomi.vim.vm.device.VirtualDisk):
				device = VirtualDisk(self._vm, key)
			elif isinstance(obj, pyVmomi.vim.vm.device.VirtualCdrom):
				device = VirtualCdrom(self._vm, key)
			elif isinstance(obj, pyVmomi.vim.vm.device.VirtualFloppy):
				device = VirtualFloppy(self._vm, key)
			elif isinstance(obj, pyVmomi.vim.vm.device.VirtualVideoCard):
				device = VirtualVideoCard(self._vm, key)
			else:
				device = VirtualDevice(self._vm, key)
			devices.append(device)
		return iter(devices)

	def __getitem__(self, index) -> 'VirtualDevice':
		return self.items[index]

	@property
	def items(self) -> 'VirtualDevice':
		"""
		A list of all items
		"""
		return list(self)

	def _add_device(self, device_spec) -> 'VirtualDevice':
		"""
		Add a new device using the provided VirtualDeviceSpec. 
		This function provides additional functionality to attempt to retrieve the added device after adding.
		Note: This function is not thread-safe as it may retrieve another device of the same type if another was simultaneously added elsewhere.

		:param device_spec: A `VirtualDeviceSpec` for the device to add.

		:return: A `VirtualDevice` (or subclass) for the added device. 
		"""
		dtype = type(device_spec.device)
		current_device_keys = [ obj.key for obj in self._vim_vm.config.hardware.device if isinstance(obj, dtype) ]
		spec = pyVmomi.vim.vm.ConfigSpec()
		spec.deviceChange = [device_spec]
		self._vm._client._wait_for_task(self._vim_vm.ReconfigVM_Task(spec=spec))
		added_device_keys = [ obj.key for obj in self._vim_vm.config.hardware.device if isinstance(obj, dtype) and obj.key not in current_device_keys ]
		added_devices = [ device for device in self if device.key in added_device_keys ]
		if len(added_devices) == 0:
			raise exceptions.VirtualMachineHardwareNotFoundError(self._vm, str(dtype), f"Failed to find newly added device")
		if len(added_devices) > 1:
			raise exceptions.VirtualMachineHardwareNotFoundError(self._vm, str(dtype), f"multiple added devices found (another device may have been added simultaneously in another thread)")
		return added_devices[0]

	def find_type(self, dtype: str) -> typing.List['VirtualDevice']:
		"""
		Find devices by their type. The type can be provided as its real value (e.g. "VirtualDisk") or as its shorthand (e.g. "disk")

		:param dtype: The type of device to get (e.g. IDEController).

		:return: List of `VirtualDevice` objects for each device matching the provided type.
		"""
		found = []
		for device in self:
			device_dtype = device._dtype.lower()
			device_dtype_shorthand = device_dtype.replace("virtual", "", 1)
			if dtype.lower() == device_dtype or dtype.lower() == device_dtype_shorthand:
				found.append(device)
		return found

	def __str__(self):
		return f"<{type(self).__name__} for {self._vm.name} ({len(list(self))} devices)>"

	def __repr__(self):
		return str(self)
		

class VirtualDevice:
	"""
	Represents a generic device on a VirtualMachine.
	"""
	def __init__(self, vm: 'VirtualMachine', key: int):
		self._vm = vm
		self._key = key

	@property
	def key(self) -> int:
		"""
		A unique key that distinguishes this device from other devices in the same virtual machine. 
		Keys are immutable but may be recycled; that is, a key does not change as long as the device is associated with a particular virtual machine. 
		However, once a device is removed, its key may be used when another device is added. 
		"""
		return self._key

	@property
	def label(self) -> str:
		"""
		The device's label
		"""
		if self._obj.deviceInfo and hasattr(self._obj.deviceInfo, "label"):
			return self._obj.deviceInfo.label.strip()
		return ""

	@property
	def connectable(self) -> bool:
		"""
		Whether or not this device is connectable. Certain functions will fail if a device is not connectable.
		"""
		return self._obj.connectable is not None

	def _assert_connectable(self):
		if not self.connectable:
			raise exceptions.VirtualMachineHardwareNotConnectableError(self._vm, self)

	@property
	def start_connected(self) -> bool:
		"""
		Whether or not this device is set to start connected.
		"""
		self._assert_connectable()
		return self._obj.connectable.startConnected

	@start_connected.setter
	def start_connected(self, value: bool):
		"""
		Set whether or not this device should start connected.
		"""
		self._assert_connectable()
		log.info(f"{str(self)} Setting device start-connected to \"{value}\"")
		device_spec = pyVmomi.vim.vm.device.VirtualDeviceSpec()
		device_spec.operation = pyVmomi.vim.vm.device.VirtualDeviceSpec.Operation.edit
		device_spec.device = self._obj
		device_spec.device.connectable.startConnected = value
		spec = pyVmomi.vim.vm.ConfigSpec()
		spec.deviceChange = [device_spec]
		self._vm._client._wait_for_task(
			self._vm._vim_vm.ReconfigVM_Task(spec=spec)
		)

	@property
	def connected(self) -> bool:
		"""
		Whether or not this device is connected.
		"""
		self._assert_connectable()
		return self._obj.connectable.connected

	@connected.setter
	def connected(self, value: bool):
		"""
		Connect or disconnect this device.
		"""
		self._assert_connectable()
		log.info(f"{str(self)} Setting device connected to \"{value}\"")
		device_spec = pyVmomi.vim.vm.device.VirtualDeviceSpec()
		device_spec.operation = pyVmomi.vim.vm.device.VirtualDeviceSpec.Operation.edit
		device_spec.device = self._obj
		device_spec.device.connectable.connected = value
		spec = pyVmomi.vim.vm.ConfigSpec()
		spec.deviceChange = [device_spec]
		self._vm._client._wait_for_task(
			self._vm._vim_vm.ReconfigVM_Task(spec=spec)
		)

	def remove(self):
		"""
		Remove this device.
		"""
		log.info(f"{str(self)} Removing device")
		device_spec = pyVmomi.vim.vm.device.VirtualDeviceSpec()
		device_spec.operation = pyVmomi.vim.vm.device.VirtualDeviceSpec.Operation.remove
		device_spec.device = self._obj
		spec = pyVmomi.vim.vm.ConfigSpec()
		spec.deviceChange = [device_spec]
		self._vm._client._wait_for_task(
			self._vm._vim_vm.ReconfigVM_Task(spec=spec)
		)

	def _get_backing_file(self) -> typing.Union['DatastoreFile', None]:
		"""
		Get the full path to the backing file.

		:return: A `DatastoreFile` object, or `None` if it does not exist.
		"""
		if self._obj.backing is None or not hasattr(self._obj.backing, "fileName"):
			return None
		return DatastoreFile.parse(self._vm._client, self._obj.backing.fileName)
	
	@property
	def _obj(self):
		for dev in self._vm._vim_vm.config.hardware.device:
			if dev.key == self.key:
				return dev
		raise exceptions.VirtualMachineHardwareNotFoundError(self._vm, type(self).__name__, f"No device with key {self.key}")

	@property
	def _dtype(self):
		return self._obj._wsdlName

	def __str__(self):
		return f"<{type(self).__name__}(dtype={self._dtype}, label=\"{self.label}\") for VM=\"{self._vm.name}\">"

	def __repr__(self):
		return str(self)


class ESXiUtilsException(Exception):
	"""Base class for all esxi_util library exceptions."""
	pass


class VirtualMachineNotFoundError(ESXiUtilsException):
	"""
	Raised when a virtual machine is not found.
	
	:param name_or_id: The name (string) or id (int) of the virtual machine
	"""
	def __init__(self, name_or_id):
		super().__init__(f"Virtual machine with {'ID' if isinstance(name_or_id, int) else 'name'} \"{name_or_id}\" not found")


class MultipleVirtualMachinesFoundError(ESXiUtilsException):
	"""
	Raised when multiple virtual machines are found but only one was expected.
	
	:param name_or_id: The name (string) or id (int) of the virtual machine
	"""
	def __init__(self, name_or_id):
		super().__init__(f"Multiple virtual machines with {'ID' if isinstance(name_or_id, int) else 'name'} \"{name_or_id}\" found")


class VirtualMachineExistsError(ESXiUtilsException):
	"""
	Raised when a virtual machine already exists.
	
	:param name: The name of the virtual machine/
	"""
	def __init__(self, name):
		super().__init__(f"A virtual machine with name \"{name}\" already exists")


class SnapshotNotFoundError(ESXiUtilsException):
	"""
	Raised when a virtual machine snapshot is not found
	
	:param vm: The VirtualMachine object
	:param name_or_id: The name (string) or id (int) of the snapshot
	"""
	def __init__(self, vm, name_or_id):
		super().__init__(f"Unable to find snapshot with {'ID' if isinstance(name_or_id, int) else 'name'} \"{name_or_id}\" for virtual machine {str(vm)}")


class MultipleSnapshotsFoundError(ESXiUtilsException):
	"""
	Raised when multiple virtual machine snapshots are found but only one was expected
	
	:param vm: The VirtualMachine object
	:param name_or_id: The name (string) or id (int) of the snapshot
	"""
	def __init__(self, vm, name_or_id):
		super().__init__(f"Multiple snapshots with {'ID' if isinstance(name_or_id, int) else 'name'} \"{name_or_id}\" found for virtual machine {str(vm)}")


class SnapshotsExistError(ESXiUtilsException):
	"""
	Raised when a virtual machine has snapshots but was expected not to have any.

	:param vm: The VirtualMachine object.
	"""
	def __init__(self, vm):
		super().__init__(f"Virtual machine {str(vm)} has snapshots")


class DatastoreError(ESXiUtilsException):
	"""
	Raised when an error is experienced related to a datastore.

	:param datastore: The datastore.
	:param reason: The reason for the error.
	"""
	def __init__(self, datastore, reason):
		super().__init__(f"Error for datastore {str(datastore)}: {str(reason)}")


class DatastoreNotFoundError(DatastoreError):
	"""
	Raised when an a datastore is not found.

	:param datastore: The datastore name.
	"""
	def __init__(self, datastore):
		super().__init__(datastore, "Not found")


class DatastoreSpaceError(DatastoreError):
	"""
	Raised when a datastore has insufficient space to complete an operation.
	
	:param datastore: The datastore.
	:param required: The space required.
	:param available: The space available.
	"""
	def __init__(self, datastore, required, available):
		super().__init__(datastore, f"Insufficient free space to complete this operation (required: {required}, available: {available})")


class DatastoreFileNotFoundError(DatastoreError):
	"""
	Raised when an a datastore file or directory is not found.

	:param datastore: The datastore.
	:param path: The path to the file
	"""
	def __init__(self, datastore, path):
		super().__init__(datastore, f"File or directory {path} not found")


class DatastoreFileExistsError(DatastoreError):
	"""
	Raised when a datastore file or directory exists but was expected not to exist.

	:param datastore: The datastore.
	:param path: The path to the file or directory.
	"""
	def __init__(self, datastore, path):
		super().__init__(datastore, f"File or directory {path} already exists")


class DatastoreNotADirectoryError(DatastoreError):
	"""
	Raised when a datastore path is not a directory when it was expected to be.

	:param datastore: The datastore.
	:param path: The path to the directory.
	"""
	def __init__(self, datastore, path):
		super().__init__(datastore, f"Path {path} is not a directory")


class DatastoreIsADirectoryError(DatastoreError):
	"""
	Raised when a path is a directory when it was expected not to be.

	:param datastore: The datastore.
	:param path: The path to the directory.
	"""
	def __init__(self, datastore, path):
		super().__init__(datastore, f"Path {path} is a directory")


class OvfImportError(ESXiUtilsException):
	"""
	Raised when an OVF or OVA has failed to import.
	
	:param path: The path to the OVF/OVA file.
	:param datastore: The name of the datastore that the OVF/OVA is being imported into.
	:param name: The name of the virtual machine that the OVF/OVA is being imported as.
	:param reason: The reason for the failure.
	"""
	def __init__(self, path, datastore, name, reason):
		super().__init__(f"Failed to import {path} as virtual machine \"{name}\" in datastore \"{datastore}\": {reason}")


class VirtualMachineExportError(ESXiUtilsException):
	"""
	Raised when a virtual machine has failed to export as OVF/OVA.
	
	:param name: The name of the virtual machine.
	:param path: The path to the file that the virtual machine is being exported to.
	:param reason: The reason for the failure.
	"""
	def __init__(self, name, path, reason):
		super().__init__(f"Failed to export virtual machine \"{name}\" to {path}: {reason}")


class VirtualMachinePowerError(ESXiUtilsException):
	"""Raised when a virtual machine is not in the desired power state."""


class VirtualMachineNotPoweredOffError(VirtualMachinePowerError):
	"""
	Raised when a virtual machine is not powered off but was expected to be.
	
	:param name: The name of the virtual machine.
	"""
	def __init__(self, name):
		super().__init__(f"Virtual machine \"{name}\" is not powered off")


class VirtualMachineAlreadyPoweredOffError(VirtualMachinePowerError):
	"""
	Raised when a virtual machine is already powered off.
	
	:param name: The name of the virtual machine.
	"""
	def __init__(self, name):
		super().__init__(f"Virtual machine \"{name}\" is already powered off")


class VirtualMachineNotPoweredOnError(VirtualMachinePowerError):
	"""
	Raised when a virtual machine is not powered on but was expected to be.
	
	:param name: The name of the virtual machine.
	"""
	def __init__(self, name):
		super().__init__(f"Virtual machine \"{name}\" is not powered on")
	

class VirtualMachineAlreadyPoweredOnError(VirtualMachinePowerError):
	"""
	Raised when a virtual machine is already powered on.
	
	:param name: The name of the virtual machine.
	"""
	def __init__(self, name):
		super().__init__(f"Virtual machine \"{name}\" is already powered on")


class ESXiShellCommandError(ESXiUtilsException):
	"""
	Raised when an ESXi shell command, or a functionality dependent on an ESXi shell command, fails.
	
	:param cmd: The command run.
	:param reason: Reason for the failure.
	:param output: Raw output of the command.
	"""
	def __init__(self, cmd, reason, output):
		super().__init__(f"ESXi shell command failure (command: \"{cmd}\"; reason: {reason}). Output: {str(output)}")


class OvfFileError(ESXiUtilsException):
	"""
	Raised when there is a failure in an OVF/OVA file itself.

	:param path: The path to the OVF file.
	:param reason: The reason for the failure.
	"""
	def __init__(self, path, reason):
		super().__init__(f"Error for file {path}: {reason}")


class OvfManifestError(OvfFileError):
	"""Raised when the manifest fails to validate for an OVF/OVA file."""


class OvfFileNotFoundError(OvfFileError):
	"""Raised when a specified OVF/OVA file is not found or is not a valid OVF/OVA file."""


class OvfAmbiguityError(OvfFileError):
	"""Raised when there is ambiguity related to an OVF/OVA file, such as when multiple matching OVF files exist."""


class RemoteConnectionError(ESXiUtilsException):
	"""
	Raised when there is an issue related to a remote connection.

	:param connection_str: A connection string used to identify the connection being made.
	:param reason: The reason for the error.
	"""
	def __init__(self, connection_str, reason):
		super().__init__(f"Connection error for {str(connection_str)}: {str(reason)}")


class RemoteConnectionNotOpenError(RemoteConnectionError):
	"""
	Raised when a remote connection is not open.

	:param connection_str: A connection string used to identify the connection being made.
	"""
	def __init__(self, connection_str):
		super().__init__(str(connection_str), "Connection is not open")


class RemoteConnectionCommandError(RemoteConnectionError):
	"""
	Raised when a remote connection command has failed.

	:param connection_str: A connection string used to identify the connection being made.
	:param response: The `Response` object for the failed command.
	"""
	def __init__(self, connection_str, response):
		super().__init__(str(connection_str), f"Command failed: {str(response)}")


class RemoteFileNotFoundError(RemoteConnectionError):
	"""
	Raised when a remote file or directory is not found.

	:param connection_str: A connection string used to identify the connection being made.
	:param path: The path to the remote file or directory.
	"""
	def __init__(self, connection_str, path):
		super().__init__(str(connection_str), f"File or directory {path} not found")


class RemoteFileExistsError(RemoteConnectionError):
	"""
	Raised when a remote file or directory exists but was expected not to exist.

	:param connection_str: A connection string used to identify the connection being made.
	:param path: The path to the remote file or directory.
	"""
	def __init__(self, connection_str, path):
		super().__init__(str(connection_str), f"File or directory {path} already exists")


class RemoteNotADirectoryError(RemoteConnectionError):
	"""
	Raised when a path is not a directory when it was expected to be.

	:param connection_str: A connection string used to identify the connection being made.
	:param path: The path to the directory.
	"""
	def __init__(self, connection_str, path):
		super().__init__(str(connection_str), f"Path {path} is not a directory")


class RemoteIsADirectoryError(RemoteConnectionError):
	"""
	Raised when a path is a directory when it was expected not to be.

	:param connection_str: A connection string used to identify the connection being made.
	:param path: The path to the directory.
	"""
	def __init__(self, connection_str, path):
		super().__init__(str(connection_str), f"Path {path} is a directory")


class PromiseCanceledException(ESXiUtilsException):
	"""
	Raised when a canceled promise is later attempted to be resolved.
	"""
	def __init__(self):
		super().__init__("Promise was canceled.")


class PromiseResolvedException(ESXiUtilsException):
	"""
	Raised when a resolved promise is re-attempted to be resolved.
	"""
	def __init__(self):
		super().__init__("Promise was already resolved.")


class GuestToolsError(ESXiUtilsException):
	"""
	Raised when there is an issue with guest tools on a virtual machine.

	:param vm: The `VirtualMachine` object.
	:param reason: The reason for the error.
	"""
	def __init__(self, vm, reason):
		super().__init__(f"Guest tools error for {str(vm)}: {str(reason)}")


class VirtualMachineHardwareError(ESXiUtilsException):
	"""
	Raised when there is an issue related to a virtual machine's hardware.

	:param vm: The `VirtualMachine` object.
	:param reason: The reason for the error.
	"""
	def __init__(self, vm, reason):
		super().__init__(f"Virtual machine hardware error for {str(vm)}: {str(reason)}")


class VirtualMachineInvalidHardwareConfigurationError(ESXiUtilsException):
	"""
	Raised when there is an issue related to a virtual machine's hardware configuration.

	:param vm: The `VirtualMachine` object.
	:param reason: The reason for the error.
	"""
	def __init__(self, vm, reason):
		super().__init__(f"Invalid hardware configuration for {str(vm)}: {str(reason)}")


class VirtualMachineHardwareNotFoundError(ESXiUtilsException):
	"""
	Raised when there is an issue with finding a virtual machine's hardware.

	:param vm: The `VirtualMachine` object.
	:param hardware_type: The type of hardware that could not be found (e.g. CD-ROM)
	:param info: Additional information that is relevant to identifying the particular hardware searched for.
	"""
	def __init__(self, vm, hardware_type, info):
		super().__init__(f"Unable to find {hardware_type} for {str(vm)}: {str(info)}")


class VirtualMachineHardwareNotConnectableError(ESXiUtilsException):
	"""
	Raised when there is a device is not connectable but an attempt to connect/disconnect the device was made.

	:param vm: The `VirtualMachine` object.
	:param device: The device in question.
	"""
	def __init__(self, vm, device):
		super().__init__(f"Device {str(device)} on {str(vm)} is not connectable")


class VNCError(ESXiUtilsException):
	"""
	Raised when there is a error with a virtual machine's VNC connection or configuration.

	:param vm: The `VirtualMachine` object.
	:param reason: The reason for the error
	"""
	def __init__(self, vm, reason):
		super().__init__(f"VNC error for {str(vm)}: {str(reason)}")


class VNCNotEnabledError(VNCError):
	"""
	Raised when VNC was expected to be enabled on a virtual machine, but is not.

	:param vm: The `VirtualMachine` object.
	"""
	def __init__(self, vm):
		super().__init__(vm, "VNC is not enabled")


class FirewallError(ESXiUtilsException):
	"""
	Raised when there is a error with ESXi's firewall.

	:param obj: The firewall-related object for error (e.g. A ruleset object).
	:param reason: The reason for the error
	"""
	def __init__(self, obj, reason):
		super().__init__(f"ESXi firewall error for {str(obj)}: {str(reason)}")


class NetworkingError(ESXiUtilsException):
	"""
	Raised when there is an error with ESXi's networking.

	:param obj: The networking-related object for the error.
	:param reason: The reason for the error.
	"""
	def __init__(self, obj, reason):
		super().__init__(f"ESXi networking error for {str(obj)}: {str(reason)}")


class NetworkingObjectNotFoundError(NetworkingError):
	"""
	Raised when a networking object is not found

	:param obj_type: The type of the object.
	:param name: The name of the object
	"""
	def __init__(self, obj_type, name):
		super().__init__(f"{obj_type} \"{name}\"", "not found")


class ESXiAPIError(ESXiUtilsException):
	"""
	Raised when there is an error related to the underlying pyVmomi API.

	:param reason: The reason for the error.
	"""
	def __init__(self, reason):
		super().__init__(f"API error, reason: {str(reason)}")


class ESXiAPIObjectNotFoundError(ESXiAPIError):
	"""
	Raised when there is an error related to finding an object using the underlying pyVmomi API.

	:param obj_type: The type of the object to be found.
	:param reason: The reason for the error.
	"""
	def __init__(self, obj_type, reason):
		super().__init__(f"Unable to find object \"{str(obj_type)}\": {str(reason)}")


class MultipleHostSystemsFoundError(ESXiUtilsException):
	"""
	Raised when multiple 'host system' machines are found but only one was expected.
	
	:param host_systems: The available host systems to choose from as a 'child' system
	"""
	def __init__(self, host_systems):
		err_str = "Multiple host systems found but no 'child_hostname' was provided! Available child hostnames: "
		for hs in host_systems:
			err_str += hs.name + "\t"
		super().__init__(err_str)


class ChildLoginFailureError(ESXiUtilsException):
	"""
	Raised when 'child' HostSystem parameters are provided to connect to a server host under a vCenter system,
	but the login fails.
	
	:param child_hostname: The child server that login was attempted on
	:param child_username: The username on the child server that login was attempted on
	"""
	def __init__(self, child_hostname, child_username):
		err_str = f"Failed to login to child HostSystem: {child_hostname} using username: {child_username}! Check that you have a connection to the server and the login credentials are correct."
		super().__init__(err_str)


class MultipleFoldersFoundError(ESXiUtilsException):
	"""
	Raised when multiple 'folder' objects are found but only one was expected.
	
	:param folders: The available folders to choose from
	"""
	def __init__(self, folders):
		err_str = "Multiple folders found with same name but only one was expected: "
		for f in folders:
			err_str += "Name: " + f.name + " ... moId: " + f._moId + "\t"
		super().__init__(err_str)


class UsbScanCodeError(ESXiUtilsException):
	"""
	Raised when there is a error with sending keys to a virtual machine.

	:param vm: The `VirtualMachine` object.
	:param reason: The reason for the error
	"""
	def __init__(self, vm, reason):
		super().__init__(f"UsbScanCodeError error for {str(vm)}: {str(reason)}")


class UsbScanCodeModifierError(ESXiUtilsException):
	"""
	Raised on attempting an application of an invalid modifier to a USB scan code key.

	:param reason: The reason for the error
	"""
	def __init__(self, reason):
		super().__init__(f"UsbScanCodeModifierError error: {str(reason)}")


class ScreenshotError(ESXiUtilsException):
	"""
	Raised when there is a error with taking a screenshot of a virtual machine.

	:param vm: The `VirtualMachine` object.
	:param reason: The reason for the error
	"""
	def __init__(self, vm, reason):
		super().__init__(f"ScreenshotError error for {str(vm)}: {str(reason)}")

class RenameError(ESXiUtilsException):
	"""
	Raised when there is a error with renaming a virtual machine.

	:param vm: The `VirtualMachine` object.
	:param reason: The reason for the error
	"""
	def __init__(self, vm, reason):
		super().__init__(f"RenameError error for {str(vm)}: {str(reason)}")



from esxi_utils.util import log, exceptions
import pyVmomi
import typing

if typing.TYPE_CHECKING:
	from esxi_utils.client import ESXiClient
	from esxi_utils.networking.distributedportgroup import DistributedPortGroup
	from esxi_utils.networking.physicalnic import PhysicalNIC

class DistributedVSwitchList:
	"""
	The list of virtual switches on the ESXi host.
	"""
	def __init__(self, client: 'ESXiClient'):
		self._client = client

	def __iter__(self) -> typing.Iterator['DistributedVSwitch']:
		return iter([ DistributedVSwitch(self._client, vdistributedswitch.dvsName) for vdistributedswitch in self._network_system.networkInfo.proxySwitch ])

	def __getitem__(self, name) -> 'DistributedVSwitch':
		return self.get(name)

	def __contains__(self, name: str) -> bool:
		return self.exists(name)

	@property
	def items(self) -> typing.List['DistributedVSwitch']:
		"""
		A list of all items
		"""
		return list(self)

	def find(self, name: str) -> typing.Union['DistributedVSwitch', None]:
		"""
		Get a virtual switch by name.

		:param name: The name of the virtual switch.

		:return: A `DistributedVSwitch` object, or `None` if not found.
		"""
		found = [ vswitch for vswitch in self if vswitch.dvsName == name ]
		if len(found) != 1:
			return None
		return found[0]

	def get(self, name: str) -> 'DistributedVSwitch':
		"""
		Get a virtual switch by name and raise an exception if not found.

		:param name: The name of the virtual switch.

		:return: A `DistributedVSwitch` object
		"""
		vswitch = self.find(name)
		if vswitch is None:
			raise exceptions.NetworkingObjectNotFoundError('DistributedVSwitch', name)
		return vswitch

	def exists(self, name: str) -> bool:
		"""
		Check whether a virtual switch exists 

		:param name: The name of the virtual switch

		:return: Whether or not the virtual switch exists
		"""
		return self.find(name) is not None

	# def add(self, name: str, mtu: int = 1500, ports: int = 128) -> 'DistributedVSwitch':
	# 	"""
	# 	Adds a new virtual switch with the given name. The name must be unique with respect to other virtual switches on the host and is limited to 32 characters.

	# 	:param name: The name for the new virtual switch.
	# 	:param mtu: The maximum transmission unit (MTU) of the virtual switch in bytes. 
	# 	:param ports: The number of ports that this virtual switch is configured to use. The maximum value is 1024, although other constraints, such as memory limits, may establish a lower effective limit.

	# 	:return: The added `DistributedVSwitch` object.
	# 	"""
	# 	assert isinstance(name, str) and len(name) > 0, "name must be a string"
	# 	assert len(name) < 32, "name must be limited to 32 characters"
	# 	assert isinstance(mtu, int) and mtu > 0, "mtu must be a positive integer"
	# 	assert isinstance(ports, int) and ports > 0, "ports must be a positive integer"
	# 	assert ports <= 1024, "ports has a maximum value of 1024"

	# 	log.info(f"Adding DistributedVSwitch \"{name}\" (mtu={mtu}, ports={ports})")
	# 	self._network_system.AddVirtualSwitch(
	# 		vswitchName=name, 
	# 		spec=pyVmomi.vim.host.VirtualSwitch.Specification(
	# 			mtu=mtu,
	# 			numPorts=ports
	# 		)
	# 	)
	# 	return DistributedVSwitch(self._client, name)

	@property
	def _network_system(self):
		return self._client._host_system.configManager.networkSystem

	def __str__(self):
		return f"<{type(self).__name__} ({len(list(self))} Virtual Switches)>"
	
	def __repr__(self):
		return str(self)


class DistributedVSwitch:
	"""
	A DistributedVSwitch on the ESXi host.
	"""
	def __init__(self, client: 'ESXiClient', name: str):
		self._client = client
		self._name = name

	@property
	def name(self) -> str:
		"""
		The name of the virtual switch.
		"""
		return self._name
	
	@property
	def portgroups(self) -> typing.List['DistributedPortGroup']:
		"""
		The DistributedPortgroups configured associated with this distributed vswitch.
		"""
		return self.distributed_portgroups

	@property
	def distributed_portgroups(self) -> typing.List['DistributedPortGroup']:
		"""
		The DistributedPortgroups associated with this distributed vswitch.
		"""
		from esxi_utils.networking.distributedportgroup import DistributedPortGroup
		portgroups = self._vim_dvs_obj.portgroup
		return [ DistributedPortGroup(self._client, pg.name) for pg in portgroups ]

	@property
	def physicalnics(self) -> typing.List["PhysicalNIC"]:
		"""
		The set of physical network adapters associated with this distributed virtual switch. 
		"""
		from esxi_utils.networking.physicalnic import PhysicalNIC
		pnics = self._obj.pnic
		return [ PhysicalNIC(self._client, pnic.device) for pnic in self._network_system.networkInfo.pnic if pnic.key in pnics ]

	@property
	def numports(self) -> int:
		"""
		The number of ports that this distributed virtual switch currently has. 
		"""
		return self._obj.numPorts

	@property
	def numports_available(self) -> int:
		"""
		The number of ports that are available on this distributed virtual switch. 
		There are a number of networking services that utilize a port on the virtual switch and are not accounted for in the Port array of a PortGroup. 
		For example, each physical NIC attached to a virtual switch consumes one port.
		"""
		return self._obj.numPortsAvailable

	# @property
	# def configured_ports(self) -> int:
	# 	"""
	# 	The number of ports that this distributed virtual switch is configured to use.
	# 	"""
	# 	return self._obj.spec.numPorts # <- value doesn't exist in distributed

	@property
	def mtu(self) -> int:
		"""
		The maximum transmission unit (MTU) associated with this distributed virtual switch in bytes.
		"""
		return self._obj.mtu

	# @property
	# def beacon(self) -> typing.Union[int, None]:
	# 	"""
	# 	The beacon configuration to probe for the validity of a link.
	# 	If this is `None`, beacon probing is disabled.
	# 	If not `None`, this is the beacon interval (how often, in seconds, a beacon should be sent)
	# 	"""
	# 	if hasattr(self._obj.spec, "bridge") and self._obj.spec.bridge and hasattr(self._obj.spec.bridge, "beacon"):
	# 		return self._obj.spec.bridge.beacon.interval
	# 	return None

	# @property
	# def link_discovery_protocol(self) -> typing.Union[str, None]:
	# 	"""
	# 	The link discovery protocol configuration for the virtual switch.
	# 	If this is `None`, this does not have a link discovery protocol.
	# 	If not `None`, this is either `cdp` (Cisco Discovery Protocol) or `lldp` (Link Layer Discovery Protocol)
	# 	"""
	# 	if hasattr(self._obj.spec, "bridge") and self._obj.spec.bridge and hasattr(self._obj.spec.bridge, "linkDiscoveryProtocolConfig"):
	# 		return self._obj.spec.bridge.linkDiscoveryProtocolConfig.protocol
	# 	return None

	# @property
	# def link_discovery_operation(self) -> typing.Union[str, None]:
	# 	"""
	# 	The link discovery operation configuration for the virtual switch.
	# 	If this is `None`, this does not have a link discovery protocol.
	# 	If not `None`, this is one of the following:
	# 	- `advertise`: 	Sent discovery packets for the switch, but don't listen for incoming discovery packets. 
	# 	- `listen`: Listen for incoming discovery packets but don't sent discovery packet for the switch. 
	# 	- `both`: Sent discovery packets for the switch and listen for incoming discovery packets. 
	# 	- `none`: Don't listen for incoming discovery packets and don't sent discover packets for the switch either. 
	# 	"""
	# 	if hasattr(self._obj.spec, "bridge") and self._obj.spec.bridge and hasattr(self._obj.spec.bridge, "linkDiscoveryProtocolConfig"):
	# 		return self._obj.spec.bridge.linkDiscoveryProtocolConfig.operation
	# 	return None

	# def add(self, name: str, vlan: int) -> 'PortGroup':
	# 	"""
	# 	Add a port group to this virtual switch.

	# 	:param name: The name of the port group.
	# 	:param vlan: The VLAN ID for ports using this port group. Possible values: 
	# 	- A value of 0 specifies that you do not want the port group associated with a VLAN.
	# 	- A value from 1 to 4094 specifies a VLAN ID for the port group.
	# 	- A value of 4095 specifies that the port group should use trunk mode, which allows the guest operating system to manage its own VLAN tags.

	# 	:return: The added `PortGroup` object.
	# 	"""
	# 	assert isinstance(name, str) and len(name) > 0, "name must be a string"
	# 	assert isinstance(vlan, int) and vlan >= 0 and vlan <= 4095, "vlan must be a integer in range 0-4095"
	# 	from esxi_utils.networking.portgroup import PortGroup
	# 	self._network_system.AddPortGroup(
	# 		portgrp=pyVmomi.vim.host.PortGroup.Specification(
	# 			name=name,
	# 			policy=pyVmomi.vim.host.NetworkPolicy(),
	# 			vlanId=vlan,
	# 			vswitchName=self.name
	# 		)
	# 	)
	# 	return PortGroup(self._client, name)

	# def remove(self):
	# 	"""
	# 	Remove this virtual switch from the system.
	# 	"""
	# 	self._network_system.RemoveVirtualSwitch(self.name)

	@property
	def _obj(self):
		for vswitch in self._network_system.networkInfo.proxySwitch:
			if vswitch.dvsName == self.name:
				return vswitch
		raise exceptions.NetworkingObjectNotFoundError('DistributedVSwitch', self.name)
	
	@property
	def _d_obj(self):
		for vswitch in self._distributed_system:
			if vswitch.switchName == self.name:
				return vswitch
		raise exceptions.NetworkingObjectNotFoundError('DistributedVSwitch', self.name)
	
	@property
	def _vim_dvs_obj(self):
		return self._d_obj.distributedVirtualSwitch

	@property
	def _network_system(self):
		return self._client._host_system.configManager.networkSystem
	
	@property
	def _distributed_system(self):
		return self._client._host_system.parent.environmentBrowser.QueryConfigTarget().distributedVirtualSwitch

	def __str__(self):
		return f"<{type(self).__name__} \"{self.name}\" ({len(self.distributed_portgroups)} distributed port groups)>"
	
	def __repr__(self):
		return str(self)



from esxi_utils.vm.virtualmachine import VirtualMachine
from esxi_utils.vm.tools.unix import UnixGuestTools
from esxi_utils.util.connect.unix import UnixSSHConnection
import typing

if typing.TYPE_CHECKING:
	from esxi_utils.vm.types.ostype import OSType

class LinuxVirtualMachine(VirtualMachine):
	"""
	A Linux virtual machine on an ESXi host. 
	"""
	@property
	def ssh(self) -> typing.Type['UnixSSHConnection']:
		"""
		An SSH connection object for establishing a remote connection to this Unix-like VM.
		"""
		return UnixSSHConnection

	@property
	def ostype(self) -> 'OSType':
		"""
		The `OSType` for this Virtual Machine.
		"""
		from esxi_utils.vm.types.ostype import OSType
		return OSType.Linux

	@property
	def tools(self) -> 'UnixGuestTools':
		"""
		Get the VMware Tools object for this VM. Includes additional methods for interaction with a Unix or Unix-like operating system.
		"""
		return UnixGuestTools(self)


from esxi_utils.util import log, exceptions
import numpy as np
import cv2 as cv
import tempfile
import typing
import random
import time
import os


if typing.TYPE_CHECKING:
	from esxi_utils.vm.virtualmachine import VirtualMachine

class ScreenCaptureHandler:
	"""
	Handler class for capturing screens on a VirtualMachine (replaces VNC screen capture functions).

	:param vm:
		A `VirtualMachine` object
	"""
	def __init__(self, vm: 'VirtualMachine'):
		self._vm = vm
		self._client = vm._client

	def capture_screen(self, filename: str = '.', timeout: float = 30.0) -> str:
		"""
		Capture a screenshot of the VM's screen.
		Raises a 'VirtualMachineNotPoweredOnError' exception if the VM is powered off.
		Raises a 'ScreenshotError' exception if the screenshot task returns an error.

		:param filename: 
			Path to a file or directory where the screenshot should be captured.
			If a directory, the screenshot will be saved as `capture.png` in that directory.
		:param timeout:
			The maximum number of seconds to wait for the operation to complete before raising a `TimeoutError`.

		:return: The path to the saved screenshot.
		"""
		# request a screenshot of the VM be taken
		# CreateScreenshot() is also an option (not documented in API)
		screenshot_task = self._vm._vim_vm.CreateScreenshot_Task()
		# retries for MKS not available error
		retries = 0
		max_retries = 25
		# start time for generic timeout error
		start_time = time.time()
		while screenshot_task.info.state != 'success':
			s = screenshot_task.info.state
			if s == 'error':
				e = screenshot_task.info.error if screenshot_task.info.error else None
				str_e = ""
				if e:
					str_e = str(e).lower()
				screenshot_info_lower = str(screenshot_task.info).lower()
				if ( e and "mks' is not available" in str_e ) or "mks' is not available" in screenshot_info_lower or (e and "mksvmx' is not available" in str_e) or "mksvmx' is not available" in screenshot_info_lower:
					if retries >= max_retries:
						raise exceptions.ScreenshotError(self._vm, f'Failed to capture_screen (timeout)! ... task info: {screenshot_task.info}')
					log.debug(f"{str(self)} Mouse/Keyboard/Screen (MKS or MKSVMX) device not available for screenshot, retrying...")
					screenshot_task = self._vm._vim_vm.CreateScreenshot_Task()
					time.sleep(1)
					retries += 1
					continue
				elif self._vm.powered_off or ( e and "(powered off)." in str_e ) or "(powered off)." in screenshot_info_lower:
					raise exceptions.VirtualMachineNotPoweredOnError(self._vm)
				if e:
					log.error(f'Screenshot Task error returned by ESXi: {str(e)}')
				raise exceptions.ScreenshotError(self._vm, f'Failed to capture_screen (uncaught exception)! ... task info: {screenshot_task.info}')
			if s == 'queued':
				now = time.time()
				if now >= start_time + timeout:
					raise exceptions.ScreenshotError(self._vm, f'Failed to capture_screen! Timeout! Task info: {screenshot_task.info}')
			# else state is either 'running' or 'success'
		
		# result format:
		# '[datastore_name] VM_name/VM_name-3.png',
		result: str = screenshot_task.info.result

		# find the datastore name and the file path in the screenshot result
		second_bracket: int = result.find(']')
		ds_name: str = result[result.find('[')+1:second_bracket]
		remote_file_path: str = result[second_bracket+1:].strip()

		# Input validation for requested output path
		output_filepath = ScreenCaptureHandler._handle_file_pathing(filename)

		# find the datastore with the given name
		ds_obj = None
		for ds in self._vm._client.datastores:
			if ds.name == ds_name:
				ds_obj = ds
				break
		if ds_obj is None:
			raise exceptions.DatastoreNotFoundError(ds_name)
		
		# grab the DatastoreFile object that represents the created screenshot
		ds_image_file = ds_obj.root / remote_file_path
		# Download the screenshot from the datastore and save it to the requested filename
		dr = ds_image_file.download(dst=output_filepath, overwrite=True)
		# Remove the screenshot from the datastore
		ds_image_file.remove()
		# return the path to the created file
		return dr[0]

	@staticmethod
	def _handle_file_pathing(filename: str):
		"""
		Ensures that the filename provided by the user points to a valid path
		
		:param filename:
			the filename to verify/fix
		:return:
			the verified/corrected filename
		"""
		filename = os.path.abspath( os.path.expanduser(filename) )
		if os.path.isdir(filename):
			filename = os.path.join(filename, "capture.png")
		if not filename.lower().endswith(".png"):
			filename += ".png"
		if not os.path.isdir(os.path.dirname(filename)):
			raise NotADirectoryError(f"{os.path.dirname(filename)} is not an existing directory")
		return filename

	def stream_screen(self, dirname: str, timeout: float = 5, continue_on_timeout: bool = True, pause: float = 0.0, continuefunc: typing.Optional[typing.Callable[[], bool]] = None) -> typing.Generator[str, None, None]:
		"""
		Continuously capture screenshots of the VM's screen in a loop. Use a ``break`` statement, ``.close()``, or the ``continuefunc`` parameter to exit the stream.

		Example usage (loop)::

			i = 0
			for screenshot_png in vm.vnc.stream_screen("screenshots/"):
				# screenshot_png = path to the taken screenshot
				# Do something with the screenshot
				i += 1
				if i == 10:
					break
					
		Example usage (next)::

			screenshot_generator = vm.vnc.stream_screen("screenshots/")
			screenshot_1_png = next(screenshot_generator)
			screenshot_2_png = next(screenshot_generator)
			screenshot_3_png = next(screenshot_generator)
			screenshot_generator.close()

		Example usage (callback)::

			i = 0
			for screenshot_png in vm.vnc.stream_screen("screenshots/", continuefunc=lambda: i < 10):
				# screenshot_png = path to the taken screenshot
				# Do something with the screenshot
				i += 1

		:param dirname:
			The directory in which to save screenshots. Screenshots will be PNG files named with the time they are taken (nanoseconds since epoch).
		:param timeout:
			The maximum number of seconds to wait per screenshot.
		:param continue_on_timeout:
			Whether to continue taking screenshots if a timeout is reached (if ``False``, a ``TimeoutError`` is raised instead).
		:param pause:
			Number of seconds to pause between each screenshot.
		:param continuefunc:
			An optional function that is called prior to taking each screenshot. If this function returns ``False``, no more screenshots will be taken and this generator will exit.

		:yields:
			The paths to each screenshot taken.
		"""
		assert isinstance(dirname, str) and len(dirname) > 0, "dirname not specified"
		assert isinstance(timeout, (int, float)) and timeout > 0, "timeout is not a number greater than 0"
		assert isinstance(continue_on_timeout, bool), "continue_on_timeout is not boolean"
		assert isinstance(pause, (int, float)) and pause >= 0, "pause is not a number greater or equal to 0"
		assert continuefunc is None or callable(continuefunc), "continuefunc is not None or a callable"

		if not os.path.isdir(dirname):
			raise NotADirectoryError(f"{dirname} is not a directory")
		
		def vm_is_off():
			log.debug(f"{str(self)} Cannot take screenshot: VM is powered off. Waiting for power on...")
			time.sleep(10)

		while continuefunc is None or continuefunc() != False:
			if self._vm.powered_off:
				vm_is_off()
				continue
			try:
				path = os.path.join(dirname, str(time.time_ns()) + ".png")
				path = self.capture_screen(filename=path, timeout=timeout)
				yield path
			except TimeoutError as e:
				if not continue_on_timeout:
					raise e
			except exceptions.VirtualMachineNotPoweredOnError:
				vm_is_off()
				continue
			if pause:
				time.sleep(pause)

	def expect_screen(self, filename: str, timeout: int = 60, match_score: float = 0.99):
		"""
		Wait until the display matches a target image. If this fails to get a match on the display in time, a `TimeoutError` will be raised.

		If the provided image is a different size than the image obtained from the screen, scaling will be automatically applied to make the images comparible.

		:param filename: An image file to read and compare against.
		:param timeout: The maximum number of seconds to wait for the operation to complete before raising a `TimeoutError`.
		:param match_score: The minimum score in range [0.0, 1.0] for images to be considered a match (higher is a closer match). A value of 1.0 indicates an exact match.
		"""
		assert isinstance(filename, str) and len(filename) > 0, "filename not specified"

		filename = os.path.abspath(filename)
		if not os.path.isfile(filename):
			raise FileNotFoundError(f"{filename} is not a file")

		target_image = cv.imread(filename)
		if target_image is None:
			raise ValueError(f"File {filename} could not be read. Ensure that this file exists and is a valid image.")
		
		def compare(image: np.ndarray, template: np.ndarray) -> float:
			try:
				res = cv.matchTemplate(image, template, cv.TM_CCOEFF_NORMED)
				score = cv.minMaxLoc(res)[1]
				return score
			except Exception:
				return 0.0

		with tempfile.TemporaryDirectory() as tempdir:

			start_time = time.time()
			while True:
				if time.time() >= start_time + timeout:
					raise TimeoutError(f"{self}: timeout while attempts to match screen to {filename}")
				
				try:
					capture_path = self.capture_screen(os.path.join(tempdir, 'screenshot.png'))
					captured_image = cv.imread(capture_path)
					if captured_image is None:
						raise RuntimeError("imread failed")
					
					# Use OpenCV to compare the images
					if captured_image.shape == target_image.shape:
						if compare(captured_image, target_image) >= match_score:
							return
					else:
						# Compare scaled versions to handle different sizes
						scaled_target_image = cv.resize(target_image, (captured_image.shape[1], captured_image.shape[0]), interpolation=cv.INTER_LINEAR)
						if compare(captured_image, scaled_target_image) >= match_score:
							return
						
						scaled_captured_image = cv.resize(captured_image, (target_image.shape[1], target_image.shape[0]), interpolation=cv.INTER_LINEAR)
						if compare(scaled_captured_image, target_image) >= match_score:
							return
						
				except Exception as e:
					if not 'operation is not allowed in the current state' in str(e):
						raise e
				
				time.sleep(0.5 + random.random()) # Add some randomness to avoid cycles where alternating images (e.g. blinking cursor) synchronizes with the screenshots
			
	def __str__(self):
		return f"<{type(self).__name__} for {self._vm.name}>"
	
	def __repr__(self):
		return str(self)



from esxi_utils.vm.hardware.device import VirtualDevice, VirtualDeviceList
from esxi_utils.datastore import DatastoreFile
from esxi_utils.util import log, exceptions
import pyVmomi
import typing
import re

class VirtualCdromList(VirtualDeviceList):
	"""
	The list of all virtual CD-ROMs on a Virtual Machine.
	"""
	def __iter__(self) -> typing.Iterator['VirtualCdrom']:
		cdroms = [ dev for dev in super().__iter__() if isinstance(dev, VirtualCdrom) ]
		cdroms.sort(key=lambda x: int(re.search(r"\d+", x.label).group(0)))
		return iter(cdroms)

	def __getitem__(self, index) -> 'VirtualCdrom':
		return self.items[index]

	@property
	def items(self) -> 'VirtualCdrom':
		"""
		A list of all items
		"""
		return list(self)

	def add(self, filepath: typing.Optional['DatastoreFile'] = None) -> 'VirtualCdrom':
		"""
		Add a new CD-ROM to the VM's devices.

		:param path: A `DatastoreFile` object for a file in the datastore to add to the CD-ROM, or `None` to add no file.

		:return: The `VirtualCdrom` object for the newly added CD-ROM.
		"""
		log.info(f"{str(self)} Adding new CD-ROM with file: {filepath}")
		assert isinstance(filepath, DatastoreFile) or filepath is None, "filepath must be a DatastoreFile object or None"

		ides = VirtualDeviceList(self._vm).find_type("VirtualIDEController")
		if len(ides) == 0:
			raise exceptions.VirtualMachineHardwareNotFoundError(self._vm, "IDE Controller", f"No controllers present to add CD-ROM")

		target_ide = None
		for ide in ides:
			if ide._obj.device is None or len(ide._obj.device) < 2: # Maximum of 2 CD-ROMs per IDE
				target_ide = ide
				break
		else:
			raise exceptions.VirtualMachineHardwareNotFoundError(self._vm, "IDE Controller", f"No controllers present with sufficient capacity to add CD-ROM")

		cdrom_spec = pyVmomi.vim.vm.device.VirtualDeviceSpec()
		cdrom_spec.operation = pyVmomi.vim.vm.device.VirtualDeviceSpec.Operation.add
		cdrom_spec.device = pyVmomi.vim.vm.device.VirtualCdrom()
		cdrom_spec.device.controllerKey = target_ide.key

		cdrom_spec.device.connectable = pyVmomi.vim.vm.device.VirtualDevice.ConnectInfo()
		cdrom_spec.device.connectable.allowGuestControl = True
		
		if filepath:
			cdrom_spec.device.backing = pyVmomi.vim.vm.device.VirtualCdrom.IsoBackingInfo()
			cdrom_spec.device.backing.fileName = filepath.path
			cdrom_spec.device.connectable.connected  = False
			cdrom_spec.device.connectable.startConnected = False
		else:
			cdrom_spec.device.backing = pyVmomi.vim.vm.device.VirtualCdrom.AtapiBackingInfo()
			cdrom_spec.device.backing.useAutoDetect = True
		return self._add_device(cdrom_spec)


class VirtualCdrom(VirtualDevice):
	@property
	def file(self) -> typing.Union['DatastoreFile', None]:
		"""
		The file attached to this CD-ROM as a `DatastoreFile` object.
		"""
		return self._get_backing_file()

	@file.setter
	def file(self, filepath: typing.Optional["DatastoreFile"]):
		"""
		Set the file for this CD-ROM.

		:param filepath: A `DatastoreFile` object for a file in the datastore to add to the CD-ROM, or `None` to empty the CD-ROM.
		"""
		log.info(f"{str(self)} Updating file to: {filepath}")
		assert isinstance(filepath, DatastoreFile) or filepath is None, "filepath must be a DatastoreFile object or None"
		
		cdrom_spec = pyVmomi.vim.vm.device.VirtualDeviceSpec()
		cdrom_spec.operation = pyVmomi.vim.vm.device.VirtualDeviceSpec.Operation.edit
		cdrom_spec.device = self._obj

		if filepath:
			cdrom_spec.device.backing = pyVmomi.vim.vm.device.VirtualCdrom.IsoBackingInfo()
			cdrom_spec.device.backing.fileName = filepath.path
		else:
			cdrom_spec.device.backing = pyVmomi.vim.vm.device.VirtualCdrom.AtapiBackingInfo()
			cdrom_spec.device.backing.useAutoDetect = True

		spec = pyVmomi.vim.vm.ConfigSpec()
		spec.deviceChange = [cdrom_spec]
		self._vm._client._wait_for_task(
			self._vm._vim_vm.ReconfigVM_Task(spec=spec)
		)

	def __str__(self):
		return f"<{type(self).__name__}(file='{self.file}') for VM='{self._vm.name}'>"



from esxi_utils.vm.tools.guesttools import GuestTools
from esxi_utils.vm.tools.panos import PanosGuestTools
from esxi_utils.vm.tools.unix import UnixGuestTools

__all__ = [
	"GuestTools",
	"PanosGuestTools",
	"UnixGuestTools"
]



from esxi_utils.util import log, exceptions
from abc import ABC, abstractmethod
from lxml import etree
import typing
import json

class Response:
	"""
	Encapsulates the response (result) of a command.

	:param cmd: The command that was executed.
	:param stdout: The stdout of the command.
	:param stderr: The stderr of the command.
	:param status: The return code (status) of the command.
	"""
	def __init__(self, cmd: str, stdout: typing.Union[str, bytes], stderr: typing.Union[str, bytes], status: int):
		self.cmd = cmd
		self.stdout = stdout.decode("utf-8", errors="surrogateescape").strip() if isinstance(stdout, bytes) else stdout.strip()
		self.stderr = stderr.decode("utf-8", errors="surrogateescape").strip() if isinstance(stderr, bytes) else stderr.strip()
		self.status = status
		self.ok = status == 0
		log.debug(f"New Response Object: {str(self)}")

	def xml(self) -> 'etree._Element':
		"""
		Try to parse this response's `stdout` as XML.

		:return: An `etree._Element` for the root node. Raises an exception if the `stdout` cannot be parsed as XML.
		"""
		try:
			return etree.fromstring(self.stdout)
		except ValueError:
			return etree.fromstring(str.encode(self.stdout))

	def __str__(self):
		return json.dumps(self.__dict__, indent = 4)

	def __repr__(self):
		return f"<{type(self).__name__} for \"{self.cmd}\" (status {self.status})>"


class ResponsePromise(ABC):
	"""
	Encapsulates a Promise for a Response (result of a command).
	"""
	def __init__(self):
		self._resolved = False
		self._canceled = False

	@abstractmethod
	def wait(self) -> Response:
		"""
		Block until the underlying operation provides a response.

		:return: A ``Response`` for the operation.
		"""
		if self._canceled:
			raise exceptions.PromiseCanceledException()
		if self._resolved:
			raise exceptions.PromiseResolvedException()
		self._resolved = True

	@abstractmethod
	def cancel(self):
		"""
		Cancel the underlying operation.
		"""
		if self._canceled:
			raise exceptions.PromiseCanceledException()
		self._canceled = True
